//SimpleLNS Part 1 - https://www.aeq-web.com/

#include <iostream>
#include <string>
#include <cstring>
#include <iomanip>
#include <chrono>
#include <ctime>
#include <algorithm>

#ifdef _WIN32
    #include <winsock2.h>
    #include <iphlpapi.h>
    #pragma comment(lib, "iphlpapi.lib")
    #pragma comment(lib, "Ws2_32.lib")
    #include <ws2tcpip.h>
#else
    #include <sys/types.h>
    #include <sys/socket.h>
    #include <ifaddrs.h>
    #include <netinet/in.h>
    #include <arpa/inet.h>
    #include <unistd.h>
#endif

bool initializeNetwork() {
#ifdef _WIN32
    WSADATA wsaData;
    if (WSAStartup(MAKEWORD(2, 2), &wsaData) != 0) {
        std::cerr << "WinSock could not be initialized." << std::endl;
        std::cin.get(); // Wait for input to keep window open on error
        return false;
    }
    std::cout << "WinSock successfully initialized." << std::endl;
#endif
    return true;
}

void printIPAddresses() {
#ifdef _WIN32
    // Retrieve adapter information on Windows
    ULONG bufferSize = 15000;
    PIP_ADAPTER_ADDRESSES adapterAddresses = (IP_ADAPTER_ADDRESSES*)malloc(bufferSize);
    if (GetAdaptersAddresses(AF_INET, GAA_FLAG_INCLUDE_PREFIX, nullptr, adapterAddresses, &bufferSize) == NO_ERROR) {
        for (PIP_ADAPTER_ADDRESSES adapter = adapterAddresses; adapter != nullptr; adapter = adapter->Next) {
            if (adapter->OperStatus == IfOperStatusUp) {
                for (PIP_ADAPTER_UNICAST_ADDRESS ua = adapter->FirstUnicastAddress; ua != nullptr; ua = ua->Next) {
                    char ipAddress[INET_ADDRSTRLEN];
                    sockaddr_in* sockaddr = (sockaddr_in*)ua->Address.lpSockaddr;
                    inet_ntop(AF_INET, &sockaddr->sin_addr, ipAddress, INET_ADDRSTRLEN);
                    if (std::string(ipAddress) != "127.0.0.1") {
                        std::cout << "IP Address: " << ipAddress << std::endl;
                    }
                }
            }
        }
    } else {
        std::cerr << "Could not retrieve adapter information." << std::endl;
        std::cin.get(); // Wait for input to keep window open on error
    }
    free(adapterAddresses);
    WSACleanup();
#else
    struct ifaddrs *interfaces = nullptr;
    struct ifaddrs *ifa = nullptr;
    if (getifaddrs(&interfaces) == -1) {
        std::cerr << "Could not retrieve network adapters." << std::endl;
        std::cin.get(); // Wait for input to keep window open on error
        return;
    }
    for (ifa = interfaces; ifa != nullptr; ifa = ifa->ifa_next) {
        if (ifa->ifa_addr && ifa->ifa_addr->sa_family == AF_INET) {
            char ipAddress[INET_ADDRSTRLEN];
            void* addr = &((struct sockaddr_in*)ifa->ifa_addr)->sin_addr;
            inet_ntop(AF_INET, addr, ipAddress, INET_ADDRSTRLEN);
            if (std::string(ipAddress) != "127.0.0.1") {
                std::cout << "IP Address: " << ipAddress << std::endl;
            }
        }
    }
    freeifaddrs(interfaces);
#endif
}

void processReceivedMessage(const char* buffer, int length) {
    // Get current timestamp
    auto now = std::chrono::system_clock::now();
    std::time_t currentTime = std::chrono::system_clock::to_time_t(now);
    std::tm* localTime = std::localtime(&currentTime);

    // Output RX with timestamp
    std::cout << "" << std::put_time(localTime, "%Y-%m-%d %H:%M:%S") << " - RX: ";

    // Output the first 12 bytes in hex format
    int hexLength = std::min(length, 12); // Display first 12 bytes in hex
    for (int i = 0; i < hexLength; ++i) {
        std::cout << std::hex << std::setw(2) << std::setfill('0') << (int)(unsigned char)buffer[i] << " ";
    }
    std::cout << std::endl;

    // Output the remaining data as a string (plaintext)
    if (length > 12) {
        std::cout << "--> ";
        std::cout.write(buffer + 12, length - 12);
        std::cout << std::endl;
    }
}

void startUDPListener() {
    // Initialize WinSock for Windows
#ifdef _WIN32
    WSADATA wsaData;
    if (WSAStartup(MAKEWORD(2, 2), &wsaData) != 0) {
        std::cerr << "WinSock initialization failed. Error Code: " << WSAGetLastError() << std::endl;
        std::cin.get();
        return;
    }
#endif

    // Create UDP-Socket
#ifdef _WIN32
    SOCKET udpSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (udpSocket == INVALID_SOCKET) {
        std::cerr << "Error creating socket. Code: " << WSAGetLastError() << std::endl;
        std::cin.get();
        WSACleanup();
        return;
    }
#else
    int udpSocket = socket(AF_INET, SOCK_DGRAM, 0);
    if (udpSocket < 0) {
        std::cerr << "Error creating socket." << std::endl;
        return;
    }
#endif

    // Configure server address
    sockaddr_in serverAddr;
    std::memset(&serverAddr, 0, sizeof(serverAddr));
    serverAddr.sin_family = AF_INET;
    serverAddr.sin_addr.s_addr = INADDR_ANY;
    serverAddr.sin_port = htons(1700);

    // Bind the socket to port 1700
#ifdef _WIN32
    if (bind(udpSocket, (sockaddr*)&serverAddr, sizeof(serverAddr)) == SOCKET_ERROR) {
        std::cerr << "Error binding socket to port 1700. Code: " << WSAGetLastError() << std::endl;
        std::cin.get();
        closesocket(udpSocket);
        WSACleanup();
        return;
    }
#else
    if (bind(udpSocket, (sockaddr*)&serverAddr, sizeof(serverAddr)) < 0) {
        std::cerr << "Error binding socket to port 1700." << std::endl;
        close(udpSocket);
        return;
    }
#endif

    std::cout << "UDP Listener started. Waiting for packets on port 1700..." << std::endl;

    // Buffer and client address for incoming packets
    char buffer[1024];
    sockaddr_in clientAddr;
    socklen_t clientAddrLen = sizeof(clientAddr);

    // Loop for receiving UDP packets
    while (true) {
        int recvLen = recvfrom(udpSocket, buffer, sizeof(buffer), 0, (sockaddr*)&clientAddr, &clientAddrLen);
        if (recvLen < 0) {
            std::cerr << "Error receiving packet." << std::endl;
            continue;
        }
        processReceivedMessage(buffer, recvLen);
    }

    // Close the socket
#ifdef _WIN32
    closesocket(udpSocket);
    WSACleanup();
#else
    close(udpSocket);
#endif
}

int main() {
    std::cout << "SimpleLNS V 1.0" << std::endl;

    if (!initializeNetwork()) {
        std::cerr << "Network access not available." << std::endl;
        std::cin.get();
        return 1;
    }

    std::cout << "Network access successful. Retrieving IPs:" << std::endl;
    printIPAddresses();

    // Start UDP Listener
    startUDPListener();

    return 0;
}
